@@ -55,12 +55,7 @@ static inline int unpack_callback_uint8(unpack_user* u, uint8_t d, msgpack_unpac
 
 static inline int unpack_callback_uint32(unpack_user* u, uint32_t d, msgpack_unpack_object* o)
 {
-    PyObject *p;
-    if (d > LONG_MAX) {
-        p = PyLong_FromUnsignedLong((unsigned long)d);
-    } else {
-        p = PyInt_FromLong((long)d);
-    }
+    PyObject *p = PyInt_FromSize_t((size_t)d);
     if (!p)
         return -1;
     *o = p;
